package ru.dbotthepony.kstarbound.world.api

import com.github.benmanes.caffeine.cache.Interner
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.util.HashTableInterner
import java.io.DataInputStream
import java.io.DataOutputStream

sealed class AbstractTileState {
	abstract val material: Registry.Entry<TileDefinition>
	abstract val modifier: Registry.Entry<MaterialModifier>?
	abstract val color: TileColor
	abstract val hueShift: Float
	abstract val modifierHueShift: Float

	abstract fun immutable(): ImmutableTileState
	abstract fun mutable(): MutableTileState

	fun write(stream: DataOutputStream) {
		stream.writeShort(material.id ?: 0)
		stream.writeBoolean(modifier != null)
		stream.writeShort(modifier?.id ?: 0)
		stream.writeByte(color.ordinal)
		stream.write((hueShift / 360f * 255).toInt())
		stream.write((modifierHueShift / 360f * 255).toInt())
	}

	companion object {
		fun skip(stream: DataInputStream) {
			stream.skipNBytes(2 + 1 + 1 + 2 + 1)
		}

		@JvmStatic
		protected val POOL: Interner<ImmutableTileState> = if (Starbound.DEDUP_CELL_STATES) Starbound.interner() else Interner { it }

		val EMPTY: ImmutableTileState = POOL.intern(ImmutableTileState(BuiltinMetaMaterials.EMPTY))
		val NULL: ImmutableTileState = POOL.intern(ImmutableTileState(BuiltinMetaMaterials.NULL))
	}
}
